/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

/*  ----- from tinyproxy for jftpgw
 *
 * Logs the various messages which tinyproxy produces to either a log file or
 * the syslog daemon. Not much to it...
 *
 * Copyright (C) 1998  Steven Young
 * Copyright (C) 1999  Robert James Kaes (rjkaes@flarenet.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * log.c - For the manipulation of log files.
 */

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>

#include "jftpgw.h"
#include "log.h"

#define LENGTH 64
#define LOGSIZE 800

#ifndef HAVE_SNPRINTF
#   include "snprintf.c"
#else
#   ifndef HAVE_VSNPRINTF
#     include "snprintf.c"
#   endif
#endif

extern struct configstruct config;
extern int mode_inetd;

/*
 * This routine logs messages to either the log file or the syslog function.
 */
void log(int level, const char *fmt, ...)
{
	va_list args;
	time_t nowtime;
	FILE *cf;

	static char time_string[LENGTH];
	static char str[LOGSIZE];

	if (level > config.debuglevel) {
		return;
	}

	/* we may dump the log to stderr if we have not yet
	 * opened a log file or the syslog */
	/* but don't log if we are run by inetd */
	if (mode_inetd && !config.logf) {
		return;
	}

	va_start(args, fmt);
	if (!config.syslog) {
		/* log via files */
		nowtime = time(NULL);
		/* Format is month day hour:minute:second (24 time) */
		strftime(time_string, LENGTH, "%b %d %H:%M:%S", localtime(&nowtime));
		if (!(cf = config.logf)) {
			cf = stderr;
		}

		fprintf(cf, "%s [%ld]: ", time_string, (long int) getpid());
		vfprintf(cf, fmt, args);
		fprintf(cf, "\n");
		fflush(cf);
	} else {
		int logtype = LOG_DEBUG;
		if (level < 8) {
			logtype = LOG_INFO;
		}
		if (level < 6) {
			logtype = LOG_WARNING;
		}
		if (level < 4) {
			logtype = LOG_ERR;
		}
		vsnprintf(str, LOGSIZE - 1, fmt, args);
		syslog(logtype, str);
	}

	va_end(args);
}




#define NUMBER_BUFFER   30
char* conv_int2char(signed int i) {
	char* s = (char*) malloc( NUMBER_BUFFER );
	enough_mem(s);
	snprintf(s, NUMBER_BUFFER, "%d", i);
	return s;
}

char* conv_uint2char(unsigned int i) {
	return conv_int2char(i);
}

char* conv_lint2char(signed long int i) {
	char* s = (char*) malloc( NUMBER_BUFFER );
	enough_mem(s);
	snprintf(s, NUMBER_BUFFER, "%ld", i);
	return s;
}

char* conv_luint2char(unsigned long int i) {
	return conv_lint2char(i);
}

const char* basename(const char* s) {
	const char* r, *t;

	t = r = s;
	while ((t = strchr(t, '/'))) {
		t++;
		r = t;
	}
	return r;
}


#define LOG_REPLACE_STRING       1
#define LOG_REPLACE_CHAR         2
#define LOG_REPLACE_UINT         3
#define LOG_REPLACE_LUINT        4
#define LOG_REPLACE_INT          5
#define LOG_REPLACE_LINT         6

char* log_replace_char(const char pattern, struct loginfo* li) {
	union {
		const char* replace_str;
		char replace_char;
		unsigned int replace_uint;
		unsigned long int replace_luint;
		signed int replace_int;
		signed long int replace_lint;
	} replace_val;
	int replace_type;

	switch (pattern) {
		case 'c': /* complete */
			replace_val.replace_char = li->complete ? 'c' : 'i';
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'T': /* Time taken to transmit/receive file, in seconds */
			replace_val.replace_uint = li->transfer_duration;
			replace_type = LOG_REPLACE_UINT;
			break;
		case 'b': /* Bytes sent for request */
			replace_val.replace_luint = li->transferred;
			replace_type = LOG_REPLACE_LUINT;
			break;
		case 'f': /* Filename stored or retrieved, absolute path */
			replace_val.replace_str = li->filename;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'F':
			/* Filename stored or retrieved, as the client sees
			 * it basename is just a pointer within li->filename
			 * */
			replace_val.replace_str = basename(li->filename);
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'm': /* Command (method) name received from client,
			     e.g., RETR */
			replace_val.replace_str = li->method;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'r': /* full commandline */
			replace_val.replace_str = li->cmd;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'P': /* pid */
			replace_val.replace_uint = getpid();
			replace_type = LOG_REPLACE_UINT;
			break;
		case 's': /*  Numeric FTP response code (status) */
			replace_val.replace_int = li->respcode;
			replace_type = LOG_REPLACE_INT;
			break;
		case 'y': /* tYpe */
			replace_val.replace_char = li->type;
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'w':  /* direction */
			replace_val.replace_char = li->direction;
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'o':  /* anonymous? */
			replace_val.replace_char =
				   strcmp(li->userlogin, "anonymous") == 0
				|| strcmp(li->userlogin, "ftp") == 0 ? 'a':'r';
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'e': /* sErvice */
			replace_val.replace_str = li->service;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'n': /* aNon-user */
			replace_val.replace_str = li->anon_user;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'H': /* Server host name */
			replace_val.replace_str = li->svrname;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'A': /* Server host IP */
			replace_val.replace_str = li->svrip;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'd': /*  Server host name as specified in the login  */
			replace_val.replace_str = li->svrlogin;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'h': /* Client host name */
			replace_val.replace_str = li->clntname;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'a': /* Client host IP */
			replace_val.replace_str = li->clntip;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'I': /* Server host name */
			replace_val.replace_str = li->ifipsvr;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'i': /* Client host IP */
			replace_val.replace_str = li->ifipclnt;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'l': /* Server user name (login) */
			replace_val.replace_str = li->userlogin;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'L': /* Effective server user name */
			replace_val.replace_str = li->usereffective;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'C': /* Forwarded server user name */
			replace_val.replace_str = li->userforwarded;
			replace_type = LOG_REPLACE_STRING;
			break;
		case '%': /* percent sign */
			replace_val.replace_char = '%';
			replace_type = LOG_REPLACE_CHAR;
			break;
		default:
			return (char*) 0;
	}

	switch(replace_type) {
		char* s;

		case LOG_REPLACE_STRING:
			if (replace_val.replace_str) {
				s = strdup(replace_val.replace_str);
				enough_mem(s);
			} else {
				s = strdup("");
				enough_mem(s);
			}
			return s;
		case LOG_REPLACE_CHAR:
			s = (char*) malloc(2);
			enough_mem(s);
			s[0] = replace_val.replace_char;
			s[1] = '\0';
			return s;
		case LOG_REPLACE_INT:
			return conv_int2char(replace_val.replace_int);
		case LOG_REPLACE_UINT:
			return conv_uint2char(replace_val.replace_uint);
		case LOG_REPLACE_LINT:
			return conv_lint2char(replace_val.replace_lint);
		case LOG_REPLACE_LUINT:
			return conv_luint2char(replace_val.replace_luint);
		default:
			return (char*) 0;
	}
	return (char*) 0;
}

char* log_replace_line(const char* line, struct loginfo* li) {

	char* replaced_line = (char*) 0;
	char* fragment;
	char* insert;
	int offset = 0;

	/* split up line */

	/* line:  blabla %u blubb %h bla */

	if (line[0] == '%') {
		replaced_line = log_replace_char(line[1], li);
		offset = 2;
	}
	while ((fragment = quotstrtok(line, "%", &offset))) {
		if ( ! replaced_line ) {
			replaced_line = fragment;
		} else {
			replaced_line = realloc(replaced_line,
				strlen(replaced_line) + strlen(fragment) + 1);
			enough_mem(replaced_line);
			strcat(replaced_line, fragment);
		}
		/* the key is at line[offset+1]; */
		offset++;
		insert = log_replace_char(line[offset], li);
		if ( ! insert ) {
			insert = strdup("<%x not found>");
			enough_mem(insert);
			/* replace `x' above */
			insert[2] = line[offset];
		}
		offset++;
		replaced_line = realloc(replaced_line,
				strlen(replaced_line) + strlen(insert) + 1);
		enough_mem(replaced_line);
		strcat(replaced_line, insert);
		free(insert);
	}
	return replaced_line;
}



void log_cmd_ent(struct cmdlogent_t* lent, struct loginfo* li) {

	char* commandpattern;
	size_t commandpatternsize;
	static char time_string[LENGTH];
	time_t nowtime;
	char* ws;

	commandpatternsize = strlen(li->cmd) + 3;
	commandpattern = (char*) malloc(commandpatternsize);
	enough_mem(commandpattern);

	if ((ws = strpbrk(li->cmd, " \t")) == NULL) {
		/* Command consisting of a single word */
		snprintf(commandpattern, commandpatternsize, " %s ", li->cmd);
	} else {
		commandpattern[0] = ' ';
		strncpy(&commandpattern[1], li->cmd, ws - li->cmd);
		commandpattern[ws-li->cmd+1] = ' ';
		commandpattern[ws-li->cmd+2] = '\0';
	}
	toupstr(commandpattern);

	if (strstr(lent->specs, commandpattern)) {
		/* found, log it */
		nowtime = time(NULL);

		if (strcmp(lent->style, "commonlog") == 0) {
			/* [18/Feb/2001:11:42:46 CET] */
			strftime(time_string, LENGTH,
				"[%d/%b/%Y:%H:%M:%S %Z]",
				localtime(&nowtime));
			if (checkbegin(li->cmd, "PASS")) {
				fprintf(lent->logf, "\"PASS *\"");
			} else {
				fprintf(lent->logf, "\"%s\"", li->cmd);
			}
			fprintf(lent->logf, " %d %ld\n",
					li->respcode, li->transferred);
		} else if (strcmp(lent->style, "xferlog") == 0) {
			char* replaced;
			const char* line_pattern =
				"%d %h %t %f %y %w _ %n %l %s 0 * %c";

			/* Wed Feb 14 01:41:28 2001 */
			strftime(time_string, LENGTH,
				"%a %b %d %H:%M:%S %Y",
				localtime(&nowtime));
			replaced = log_replace_line(line_pattern, li);
			fprintf(lent->logf, "%s %s\n", time_string, replaced);
			li->filename = (char*) 0;
		} else {
			char* line_pattern = lent->style;
			char* replaced;

			/* Wed Feb 14 01:41:28 2001 */
			strftime(time_string, LENGTH,
				"%a %b %d %H:%M:%S %Y",
				localtime(&nowtime));
			replaced = log_replace_line(line_pattern, li);
			fprintf(lent->logf, "%s %s\n", time_string, replaced);
			li->filename = (char*) 0;
		}
		fflush(lent->logf);
	}
	free(commandpattern);
}



void log_cmd(struct loginfo* li) {

	/* go through the specifications of the logfiles and write
	 * them if they match */

	struct cmdlogent_t* files = &config.cmdlogfiles;
	struct cmdlogent_t* dirs  = &config.cmdlogdirs;

	while (files && files->logf_name) {	
		log_cmd_ent(files, li);
		files = files->next;
	}

	while (dirs && dirs->logf_name) {
		log_cmd_ent(dirs, li);
		dirs = dirs->next;
	}
}

